@@ -13,11 +13,12 @@
 # limitations under the License.
 """Convert your NLP text data to arrays!"""
 
-__version__ = '0.0.13'
+__version__ = '0.1.0'
 __all__ = [
     'Sample',
     'Batch',
     'Vocab',
+    'StringStore',
     'BatchIterator',
     'ShuffleIterator',
 ]
@@ -25,4 +26,4 @@ __all__ = [
 from .batches import Batch
 from .samples import Sample
 from .iterators import BatchIterator, ShuffleIterator
-from .vocab import Vocab
+from .vocab import StringStore, Vocab
