@@ -12,58 +12,52 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from collections import Counter, OrderedDict, defaultdict
+from collections import Counter, UserDict, defaultdict
 from typing import Counter as CounterT, Dict, Iterable, Iterator, Mapping, \
-    Optional, Sequence, Set
+    MutableMapping, Optional, Sequence, Set
 
+from ordered_set import OrderedSet
 from tqdm import tqdm
 
 from .samples import FieldName, FieldValue, Sample
 
 
-class Vocab(Mapping[FieldName, Mapping[str, int]]):
-    """Namespaced vocabulary storing the mapping from field names to their actual vocabulary.
-
-    A vocabulary does not hold the str-to-int mapping directly, but rather it stores a mapping
-    from field names to the corresponding str-to-int mappings. These mappings are the actual
-    vocabulary for that particular field name. In other words, the actual vocabulary for each
-    field name is namespaced by the field name and all of them are handled this :class:`Vocab`
-    object.
-
-    Args:
-        m: Mapping from field names to its str-to-int mapping.
-    """
-
-    def __init__(self, m: Mapping[FieldName, Mapping[str, int]]) -> None:
-        self._m = m
-
-    def __len__(self) -> int:
-        return len(self._m)
-
-    def __iter__(self) -> Iterator[FieldName]:
-        return iter(self._m)
-
-    def __getitem__(self, name: FieldName) -> Mapping[str, int]:
+class Vocab(UserDict, MutableMapping[FieldName, 'StringStore']):
+    """A dictionary from field names to `StringStore` objects as the field's vocabulary."""
+    def __getitem__(self, name: FieldName) -> 'StringStore':
         try:
-            return self._m[name]
+            return super().__getitem__(name)
         except KeyError:
             raise KeyError(f"no vocabulary found for field name '{name}'")
 
-    def apply_to(self, samples: Iterable[Sample]) -> Iterable[Sample]:
-        """Apply this vocabulary to the given samples.
+    def stoi(self, samples: Iterable[Sample]) -> Iterable[Sample]:
+        """Convert strings in the given samples to integers according to this vocabulary.
 
-        Applying a vocabulary means mapping all the (nested) field values to the corresponding
-        values according to the mapping specified by the vocabulary. Field names that have
-        no entry in the vocabulary are ignored. Note that the actual application is not
-        performed until the resulting iterable is iterated over.
+        This conversion means mapping all the (nested) string field values to integers
+        according to the mapping specified by the `StringStore` object of that field.
+        Field names with no entry in the vocabulary are ignored. Note that the actual
+        conversion is lazy; it is not performed until the resulting iterable is iterated over.
 
         Args:
-            samples (~typing.Iterable[Sample]): Apply vocabulary to these samples.
+            samples (~typing.Iterable[Sample]): Samples to convert.
 
         Returns:
-            ~typing.Iterable[Sample]: Samples to which the vocabulary has been applied.
+            ~typing.Iterable[Sample]: Converted samples.
         """
-        return _VocabAppliedSamples(self, samples)
+        return map(self._apply_to_sample, samples)
+
+    def itos(self, samples: Iterable[Sample]) -> Iterable[Sample]:
+        """Convert integers in the given samples to strings according to this vocabulary.
+
+        This method is essentially the inverse of `~Vocab.stoi`.
+
+        Args:
+            samples (~typing.Iterable[Sample]): Samples to convert.
+
+        Returns:
+            ~typing.Iterable[Sample]: Converted samples.
+        """
+        return map(lambda s: self._apply_to_sample(s, index=False), samples)
 
     @classmethod
     def from_samples(
@@ -83,10 +77,10 @@ class Vocab(Mapping[FieldName, Mapping[str, int]]):
             pbar: Instance of `tqdm <https://pypi.org/project/tqdm>`_ for displaying
                 a progress bar.
             options: Mapping from field names to dictionaries to control the creation of
-                the str-to-int mapping. Recognized dictionary keys are:
+                the vocabularies. Recognized dictionary keys are:
 
                 * ``min_count`` (`int`): Exclude tokens occurring fewer than this number
-                  of times from the vocabulary (default: 2).
+                  of times from the vocabulary (default: 1).
                 * ``pad`` (`str`): String token to represent padding tokens. If ``None``,
                   no padding token is added to the vocabulary. Otherwise, it is the
                   first entry in the vocabulary (index is 0). Note that if the field has no
@@ -124,27 +118,27 @@ class Vocab(Mapping[FieldName, Mapping[str, int]]):
 
         m = {}
         for name, c in counter.items():
-            store: dict = OrderedDict()
             opts = options.get(name, {})
 
             # Padding and unknown tokens
             pad = opts.get('pad', '<pad>')
             unk = opts.get('unk', '<unk>')
+            inits = []
             if name in seqfield and pad is not None:
-                store[pad] = len(store)
+                inits.append(pad)
             if unk is not None:
-                store[unk] = len(store)
+                inits.append(unk)
 
-            min_count = opts.get('min_count', 2)
+            store = StringStore(inits, default=unk)
+
+            min_count = opts.get('min_count', 1)
             max_size = opts.get('max_size')
             n = len(store)
             for tok, freq in c.most_common():
                 if freq < min_count or (max_size is not None and len(store) - n >= max_size):
                     break
-                store[tok] = len(store)
-
-            unk_id = None if unk is None else store[unk]
-            m[name] = _StringStore(store, unk_id=unk_id)
+                store.add(tok)
+            m[name] = store
 
         return cls(m)
 
@@ -153,9 +147,7 @@ class Vocab(Mapping[FieldName, Mapping[str, int]]):
         if isinstance(val, str):
             return True
         if isinstance(val, Sequence):
-            if not val:
-                raise ValueError('field values must not be an empty sequence')
-            return cls._needs_vocab(val[0])
+            return False if not val else cls._needs_vocab(val[0])
         return False
 
     @classmethod
@@ -168,66 +160,87 @@ class Vocab(Mapping[FieldName, Mapping[str, int]]):
         for x in xs:
             yield from cls._flatten(x)
 
-    def _apply_to_sample(self, sample: Sample) -> Sample:
+    def _apply_to_sample(self, sample: Sample, index: bool = True) -> Sample:
+        fn = self._index_value if index else self._get_value
         s = {}
-        for name, val in sample.items():
+        for name, value in sample.items():
             try:
-                vb = self[name]
+                store = self[name]
             except KeyError:
-                s[name] = val
+                s[name] = value
             else:
-                s[name] = self._apply_vb_to_val(vb, val)
+                s[name] = fn(store, value)
         return s
 
     @classmethod
-    def _apply_vb_to_val(
-            cls,
-            vb: Mapping[FieldValue, FieldValue],
-            val: FieldValue,
-    ) -> FieldValue:
-        if isinstance(val, str) or not isinstance(val, Sequence):
-            try:
-                return vb[val]
-            except KeyError:
-                raise KeyError(f'value {val!r} not found in vocab')
+    def _index_value(cls, store: 'StringStore', value: FieldValue) -> FieldValue:
+        if isinstance(value, str):
+            return store.index(value)
+        if not isinstance(value, Sequence):
+            return value
 
-        return [cls._apply_vb_to_val(vb, v) for v in val]
+        return [cls._index_value(store, v) for v in value]
+
+    @classmethod
+    def _get_value(cls, store: 'StringStore', value: FieldValue) -> FieldValue:
+        if not isinstance(value, Sequence):
+            return store[value]
+        if isinstance(value, str):
+            return value
+
+        return [cls._get_value(store, v) for v in value]
 
 
-class _VocabAppliedSamples(Iterable[Sample]):
-    def __init__(self, vocab: Vocab, samples: Iterable[Sample]) -> None:
-        self._vocab = vocab
-        self._samples = samples
+class StringStore(OrderedSet):
+    """An ordered set of strings, with an optional default value for unknown strings.
 
-    def __iter__(self) -> Iterator[Sample]:
-        for s in self._samples:
-            yield self._vocab._apply_to_sample(s)
+    This class implements both `~typing.MutableSet` and `~typing.Sequence` with `str`
+    as its contents.
 
+    Example:
 
-class _StringStore(Mapping[str, int]):
-    def __init__(self, m: Mapping[str, int], unk_id: Optional[int] = None) -> None:
-        assert unk_id is None or unk_id >= 0
-        self._m = m
-        self._unk_id = unk_id
+        >>> from text2array import StringStore
+        >>> store = StringStore('abb', default='a')
+        >>> list(store)
+        ['a', 'b']
+        >>> store.add('b')
+        1
+        >>> store.add('c')
+        2
+        >>> list(store)
+        ['a', 'b', 'c']
+        >>> store.index('a')
+        0
+        >>> store.index('b')
+        1
+        >>> store.index('d')
+        0
 
-    def __len__(self) -> int:
-        return len(self._m)
+    Args:
+        initial: Initial elements of the store.
+        default: Default string as a representation of unknown strings, i.e. those that
+            do not exist in the store.
+    """
+    def __init__(
+            self,
+            initial: Optional[Sequence[str]] = None,
+            default: Optional[str] = None,
+    ) -> None:
+        super().__init__(initial)
+        self.default = default
 
-    def __iter__(self) -> Iterator[str]:
-        return iter(self._m)
-
-    def __getitem__(self, s: str) -> int:
+    def index(self, s: str) -> int:
         try:
-            return self._m[s]
+            return super().index(s)
         except KeyError:
-            if self._unk_id is not None:
-                return self._unk_id
-            raise KeyError(f"'{s}' not found in vocabulary")
-
-    def __contains__(self, s) -> bool:
-        return s in self._m
+            if self.default is not None:
+                return super().index(self.default)
+            raise ValueError(f"cannot find '{s}'")
 
     def __eq__(self, o) -> bool:
-        if not isinstance(o, _StringStore):
+        if not isinstance(o, StringStore):
             return False
-        return self._m == o._m and self._unk_id == o._unk_id
+        return self.default == o.default and super().__eq__(o)
+
+    def __repr__(self) -> str:
+        return f'{self.__class__.__name__}({list(self)!r}, default={self.default!r})'
